# coding=utf-8
from earthdiagnostics.diagnostic import Diagnostic, DiagnosticOption, DiagnosticDomainOption, DiagnosticVariableOption
from earthdiagnostics.modelingrealm import ModelingRealm


class Rewrite(Diagnostic):
    """
    Rewrites files without doing any calculations.
    Can be useful to convert units or to correct wrong metadata

    :original author: Javier Vegas-Regidor<javier.vegas@bsc.es>

    :created: July 2016

    :param data_manager: data management object
    :type data_manager: DataManager
    :param startdate: startdate
    :type startdate: str
    :param member: member number
    :type member: int
    :param chunk: chunk's number
    :type chunk: int
    :param variable: variable's name
    :type variable: str
    :param domain: variable's domain
    :type domain: ModelingRealm
    """

    alias = 'rewrite'
    "Diagnostic alias for the configuration file"

    def __init__(self, data_manager, startdate, member, chunk, domain, variable, grid):
        Diagnostic.__init__(self, data_manager)
        self.startdate = startdate
        self.member = member
        self.chunk = chunk
        self.variable = variable
        self.domain = domain
        self.grid = grid

    def __str__(self):
        return 'Rewrite output Startdate: {0} Member: {1} Chunk: {2} ' \
               'Variable: {3}:{4} Grid: {5}'.format(self.startdate, self.member, self.chunk, self.domain, self.variable,
                                                    self.grid)

    def __eq__(self, other):
        return self.startdate == other.startdate and self.member == other.member and self.chunk == other.chunk and \
            self.domain == other.domain and self.variable == other.variable and self.grid == self.grid

    @classmethod
    def generate_jobs(cls, diags, options):
        """
            Creates a job for each chunk to compute the diagnostic

            :param diags: Diagnostics manager class
            :type diags: Diags
            :param options: variable, domain, grid
            :type options: list[str]
            :return:
            """
        options_available = (DiagnosticDomainOption(),
                             DiagnosticVariableOption(),
                             DiagnosticOption('grid', ''))
        options = cls.process_options(options, options_available)
        job_list = list()
        for startdate, member, chunk in diags.config.experiment.get_chunk_list():
            job_list.append(Rewrite(diags.data_manager, startdate, member, chunk,
                                    options['domain'], options['variable'], options['grid']))
        return job_list

    def request_data(self):
        self.variable_file = self.request_chunk(self.domain, self.variable, self.startdate, self.member, self.chunk,
                                                grid=self.grid, to_modify=True)

    def declare_data_generated(self):
        self.corrected = self.declare_chunk(self.domain, self.variable, self.startdate, self.member, self.chunk,
                                            grid=self.grid)

    def compute(self):
        """
        Runs the diagnostic
        """
        self.corrected.set_local_file(self.variable_file.local_file, self)

