# coding=utf-8

import os
from earthdiagnostics.diagnostic import Diagnostic, DiagnosticOption, DiagnosticDomainOption, \
    DiagnosticFrequencyOption, DiagnosticVariableOption
from earthdiagnostics.frequency import Frequencies
from earthdiagnostics.utils import Utils, TempFile
from earthdiagnostics.modelingrealm import ModelingRealm


class YearlyMean(Diagnostic):
    """
    Calculates yearly mean for a given variable

    :original author: Javier Vegas-Regidor<javier.vegas@bsc.es>

    :created: July 2016

    :param data_manager: data management object
    :type data_manager: DataManager
    :param startdate: startdate
    :type startdate: str
    :param member: member number
    :type member: int
    :param chunk: chunk's number
    :type chunk: int
    :param variable: variable's name
    :type variable: str
    :param domain: variable's domain
    :type domain: ModelingRealm
    :param frequency: original frequency
    :type frequency: str
    :param grid: original data grid
    :type grid: str
    """

    alias = 'yearmean'
    "Diagnostic alias for the configuration file"

    def __init__(self, data_manager, startdate, member, chunk, domain, variable, frequency, grid):
        Diagnostic.__init__(self, data_manager)
        self.startdate = startdate
        self.member = member
        self.chunk = chunk
        self.variable = variable
        self.domain = domain
        self.frequency = frequency
        self.grid = grid

    def __str__(self):
        return 'Calculate yearly mean Startdate: {0} Member: {1} Chunk: {2} ' \
               'Variable: {3}:{4} Original frequency: {5} Grid: {6}'.format(self.startdate, self.member, self.chunk,
                                                                            self.domain,  self.variable,
                                                                            self.frequency, self.grid)

    def __eq__(self, other):
        return self.startdate == other.startdate and self.member == other.member and self.chunk == other.chunk and \
            self.domain == other.domain and self.variable == other.variable and self.frequency == other.frequency and \
            self.grid == other.grid

    @classmethod
    def generate_jobs(cls, diags, options):
        """
            Creates a job for each chunk to compute the diagnostic

            :param diags: Diagnostics manager class
            :type diags: Diags
            :param options: variable, domain, frequency=day, grid=''
            :type options: list[str]
            :return:
            """

        options_available = (DiagnosticDomainOption(),
                             DiagnosticVariableOption(),
                             DiagnosticFrequencyOption(),
                             DiagnosticOption('grid', ''))
        options = cls.process_options(options, options_available)
        job_list = list()
        for startdate, member, chunk in diags.config.experiment.get_chunk_list():
            job_list.append(YearlyMean(diags.data_manager, startdate, member, chunk,
                                       options['domain'], options['variable'], options['frequency'], options['grid']))
        return job_list

    def request_data(self):
        self.variable_file = self.request_chunk(self.domain, self.variable, self.startdate, self.member, self.chunk,
                                                frequency=self.frequency, grid=self.grid)

    def declare_data_generated(self):
        self.yearmean = self.declare_chunk(self.domain, self.variable, self.startdate, self.member, self.chunk,
                                           frequency=Frequencies.yearly, grid=self.grid)

    def compute(self):
        """
        Runs the diagnostic
        """
        temp = TempFile.get()

        handler = Utils.openCdf(self.variable_file.local_file)
        if 'region' in handler.variables:
            noregion = TempFile.get()
            Utils.nco.ncks(input=self.variable_file.local_file, output=noregion, options=('-O -C -x -v region',))
            Utils.cdo.yearmean(input=noregion, output=temp)
            monmean_handler = Utils.openCdf(temp)
            Utils.copy_variable(handler, monmean_handler, 'region')
            monmean_handler.close()
        else:
            Utils.cdo.yearmean(input=self.variable_file.local_file, output=temp)
        os.remove(self.variable_file.local_file)

        self.yearmean.set_local_file(temp)

